# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


def create_default_groups(apps, schema_editors):
    PERM_CONF = {
        "publish_content": "Can publish content",
        "publish_own_content": "Can publish own content",
        "change_content": "Can change content",
        "promote_content": "Can promote content"
    }
    GROUP_CONF = dict(
        contributor=(),
        author=("publish_own_content",),
        editor=(
            "publish_content",
            "change_content",
            "promote_content",
        ),
        admin=(
            "publish_content",
            "change_content",
            "promote_content",
        )
    )
    ContentType = apps.get_model("contenttypes", "ContentType")
    Group = apps.get_model("auth", "Group")
    Permission = apps.get_model("auth", "Permission")
    content_ct, _ = ContentType.objects.get_or_create(
        model="content", app_label="content"
    )
    for group_name, group_perms in GROUP_CONF.items():
        group, _ = Group.objects.get_or_create(
            name=group_name
        )
        for perm_name in group_perms:
            perm, _ = Permission.objects.get_or_create(
                content_type=content_ct,
                codename=perm_name,
                defaults={
                    "name": PERM_CONF[perm_name]
                }
            )
            group.permissions.add(perm)


class Migration(migrations.Migration):

    dependencies = [
        ('content', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_default_groups, lambda x,y: None),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import djbetty.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Campaign',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('sponsor_name', models.CharField(max_length=255)),
                ('sponsor_logo', djbetty.fields.ImageField(default=None, null=True, blank=True)),
                ('sponsor_url', models.URLField(null=True, blank=True)),
                ('start_date', models.DateTimeField(null=True, blank=True)),
                ('end_date', models.DateTimeField(null=True, blank=True)),
                ('campaign_label', models.CharField(max_length=255)),
                ('impression_goal', models.IntegerField(null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='CampaignPixel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('url', models.URLField()),
                ('campaign_type', models.IntegerField(default=0, choices=[(0, b'Logo')])),
                ('campaign', models.ForeignKey(related_name='pixels', to='campaigns.Campaign')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from django.db import models

from .models import VideohubVideo


class VideoMixin(models.Model):

    """Provides an OnionStudios (videohub) reference ID, standardized across all properties."""

    videohub_ref = models.ForeignKey(VideohubVideo, null=True, blank=True)

    class Meta:
        abstract = True



from rest_framework import serializers

from bulbs.content.models import Content
from bulbs.content.serializers import ContentSerializer, ContentTypeField

from .models import PZone
from .operations import PZoneOperation, InsertOperation, DeleteOperation, ReplaceOperation


class PZoneField(serializers.Field):
    # def field_to_native(self, obj, field_name):
    #     data = []
    #     for content in obj:
    #         data.append(ContentSerializer(instance=content).data)

    #     return data

    def to_representation(self, obj):
        data = []
        bulk_content = Content.objects.in_bulk([content_data["id"] for content_data in obj])
        for content_data in obj:
            content = bulk_content[content_data["id"]]
            data.append(ContentSerializer(instance=content).data)
        return data

    def to_internal_value(self, data):
        return [{"id": content_data["id"]} for content_data in data]


class PZoneSerializer(serializers.ModelSerializer):
    class Meta:
        model = PZone
        exclude = ("data",)

    content = PZoneField(source="data", required=False)


class _PZoneOperationSerializer(serializers.ModelSerializer):
    """Parent class of pzone operation serializers."""

    type_name = ContentTypeField(source="polymorphic_ctype_id")
    pzone = serializers.PrimaryKeyRelatedField(queryset=PZone.objects.all())
    when = serializers.DateTimeField()
    applied = serializers.BooleanField(default=False)
    content = serializers.PrimaryKeyRelatedField(queryset=Content.objects.all())
    content_title = serializers.SerializerMethodField()

    class Meta:
        model = PZoneOperation

    def get_content_title(self, obj):
        """Get content's title."""
        return Content.objects.get(id=obj.content.id).title


class InsertOperationSerializer(_PZoneOperationSerializer):

    index = serializers.IntegerField(min_value=0)

    class Meta:
        model = InsertOperation


class ReplaceOperationSerializer(_PZoneOperationSerializer):

    index = serializers.IntegerField(min_value=0)

    class Meta:
        model = ReplaceOperation


class DeleteOperationSerializer(_PZoneOperationSerializer):

    class Meta:
        model = DeleteOperation



from collections import OrderedDict

from django.utils.encoding import force_text

from rest_framework import serializers
from rest_framework.metadata import SimpleMetadata


METADATA_ATTRIBUTES = [
    "field_size", "read_only", "label", "help_text", "min_length", "max_length",
    "min_value", "max_value", "child_label"
]


class BaseSimpleMetadata(SimpleMetadata):
    custom_serializer = None
    custom_data_field = None

    def get_label_lookup(self, field):
        field_type = self.label_lookup[field]
        if field_type == "field" and hasattr(field, "child_relation"):
            return self.get_label_lookup(field.child_relation)
        return field_type

    def get_attributes(self, obj):
        info = dict()
        for attr in sorted(METADATA_ATTRIBUTES):
            value = getattr(obj, attr, None)
            if value is not None and value != "":
                info[attr] = force_text(value, strings_only=True)
        return info

    def determine_metadata(self, request, view):
        serializer_class = view.get_serializer_class()
        if issubclass(serializer_class, self.custom_serializer):
            data = self.get_custom_metadata(serializer_class, view)
            return data
        return super(BaseSimpleMetadata, self).determine_metadata(request, view)

    def get_field_info(self, field):
        """
        This method is basically a mirror from rest_framework==3.3.3

        We are currently pinned to rest_framework==3.1.1. If we upgrade,
        this can be refactored and simplified to rely more heavily on
        rest_framework's built in logic.
        """

        field_info = self.get_attributes(field)
        field_info["required"] = getattr(field, "required", False)
        field_info["type"] = self.get_label_lookup(field)

        if getattr(field, "child", None):
            field_info["child"] = self.get_field_info(field.child)
        elif getattr(field, "fields", None):
            field_info["children"] = self.get_serializer_info(field)

        if (not isinstance(field, (serializers.RelatedField, serializers.ManyRelatedField)) and
                hasattr(field, "choices")):
            field_info["choices"] = [
                {
                    "value": choice_value,
                    "display_name": force_text(choice_name, strings_only=True)
                }
                for choice_value, choice_name in field.choices.items()
            ]

        return field_info

    def get_custom_field_name(self, view):
        raise NotImplementedError

    def get_custom_metadata(self, serializer, view):
        fields_metadata = dict()
        if hasattr(serializer, "__call__"):
            serializer_instance = serializer()
        else:
            serializer_instance = serializer
        for field_name, field in serializer_instance.get_fields().items():
            if isinstance(field, self.custom_data_field):
                if view.suffix != "List":
                    fields_metadata[field_name] = self.get_custom_field_name(view)
            elif isinstance(field, serializers.BaseSerializer):
                fields_metadata[field_name] = self.get_serializer_info(field)
            else:
                fields_metadata[field_name] = self.get_field_info(field)
        return {
            "fields": fields_metadata
        }

    def get_serializer_info(self, serializer):
        serializer_info = super(BaseSimpleMetadata, self).get_serializer_info(serializer)
        custom_params = OrderedDict()

        if hasattr(serializer, "child"):
            label = self.label_lookup[serializer.child]
            if hasattr(serializer.child, "child_label"):
                custom_params["child_label"] = serializer.child.child_label
        else:
            label = self.label_lookup[serializer]

        if label != "field":
            serializer_info = OrderedDict([
                ("type", label),
                ("fields", serializer_info)
            ])
            for key, value in custom_params.items():
                serializer_info[key] = value

        return serializer_info



from django_enumfield import enum


class InfographicType(enum.Enum):
    LIST = 0
    TIMELINE = 1
    STRONGSIDE_WEAKSIDE = 2
    PRO_CON = 3
    COMPARISON = 4

    labels = {
        LIST: "List",
        TIMELINE: "Timeline",
        STRONGSIDE_WEAKSIDE: "Strongside_Weakside",
        PRO_CON: "Pro_Con",
        COMPARISON: "Comparison"
    }



from django.conf.urls import url, include

from bulbs.contributions import signals  # NOQA

from .views import api_v1_router, api_v1_role_router, nested_role_router

urlpatterns = (
    url(r"^", include(api_v1_router.urls)),
    url(r"^", include(api_v1_role_router.urls)),
    url(r"^", include(nested_role_router.urls))
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('campaigns', '0004_auto_20151103_1358'),
    ]

    operations = [
        migrations.AddField(
            model_name='campaign',
            name='tunic_campaign_id',
            field=models.IntegerField(default=None, null=True, blank=True),
        ),
    ]



# Hashicorp Vault client interface
#
# Should import module (instead of function) for easy vault test mocks:
#
#   import vault
#   vault.read(...)
#
import requests

from django.conf import settings

log = __import__('logging').getLogger(__name__)


def read(path):
    """Read a secret from Vault REST endpoint"""
    url = '{}/{}/{}'.format(settings.VAULT_BASE_URL.rstrip('/'),
                            settings.VAULT_BASE_SECRET_PATH.strip('/'),
                            path.lstrip('/'))

    headers = {'X-Vault-Token': settings.VAULT_ACCESS_TOKEN}
    resp = requests.get(url, headers=headers)
    if resp.ok:
        return resp.json()['data']
    else:
        log.error('Failed VAULT GET request: %s %s', resp.status_code, resp.text)
        raise Exception('Failed Vault GET request: {} {}'.format(resp.status_code, resp.text))



from django.template import loader


class BaseRenderer:

    def generate_body(self, intermediate):
        body = []
        for item in intermediate:
            for key, values in item.items():
                body.append(self.render_item(key, values).strip())
        return '\n'.join(body)

    def render_item(self, key, body):
        if key == "text":
            return body["raw"]
        elif key == "betty":
            return self.render(self.BETTY_TEMPLATE, body)
        elif key == "facebook":
            return self.render(self.FACEBOOK_TEMPLATE, body)
        elif key == "imgur":
            return self.render(self.IMGUR_TEMPLATE, body)
        elif key == "instagram":
            return self.render(self.INSTAGRAM_TEMPLATE, body)
        elif key == "onion_video":
            return self.render(self.ONION_VIDEO_TEMPLATE, body)
        elif key == "soundcloud":
            return self.render(self.SOUNDCLOUD_TEMPLATE, body)
        elif key == "twitter":
            return self.render(self.TWITTER_TEMPLATE, body)
        elif key == "vimeo":
            return self.render(self.VIMEO_TEMPLATE, body)
        elif key == "youtube":
            return self.render(self.YOUTUBE_TEMPLATE, body)
        else:
            raise Exception("Key not implemented")

    def render(self, template, body):
        return loader.render_to_string(
            template,
            body
        )


class InstantArticleRenderer(BaseRenderer):

    BETTY_TEMPLATE = "instant_article/embeds/_ia_betty_embed.html"
    FACEBOOK_TEMPLATE = "instant_article/embeds/_ia_facebook_embed.html"
    IMGUR_TEMPLATE = "instant_article/embeds/_ia_imgur_embed.html"
    INSTAGRAM_TEMPLATE = "instant_article/embeds/_ia_instagram_embed.html"
    ONION_VIDEO_TEMPLATE = "instant_article/embeds/_ia_onion_video_embed.html"
    SOUNDCLOUD_TEMPLATE = "instant_article/embeds/_ia_soundcloud_embed.html"
    TWITTER_TEMPLATE = "instant_article/embeds/_ia_twitter_embed.html"
    VIMEO_TEMPLATE = "instant_article/embeds/_ia_vimeo_embed.html"
    YOUTUBE_TEMPLATE = "instant_article/embeds/_ia_youtube_embed.html"



from rest_framework import filters, viewsets
from rest_framework.permissions import IsAdminUser

from .models import Notification
from .serializers import NotificationSerializer


class NotificationViewSet(viewsets.ModelViewSet):

    queryset = Notification.objects.all().order_by('-created_on')
    serializer_class = NotificationSerializer
    filter_backends = (filters.DjangoFilterBackend,
                       filters.OrderingFilter,
                       filters.SearchFilter)
    filter_fields = ('is_published',)
    search_fields = ('internal_title',)
    ordering_fields = ('internal_title',
                       'is_published',
                       'created_on')
    paginate_by = 20
    permission_classes = [IsAdminUser]


class ReadOnlyNotificationViewSet(viewsets.ReadOnlyModelViewSet):

    queryset = Notification.objects.filter(is_published=True)
    serializer_class = NotificationSerializer
    paginate_by = 20



"""Base models for "Content", including the indexing and search features
that we want any piece of content to have."""

import logging
import uuid
import requests

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import NoReverseMatch, reverse
from django.db import models
from django.db.models import Model
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.html import strip_tags

from djbetty import ImageField
from djes.models import Indexable, IndexableManager
from elasticsearch import TransportError
from elasticsearch_dsl import field
from polymorphic import PolymorphicModel, PolymorphicManager

from bulbs.content import TagCache
from bulbs.content.tasks import (
    index_content_contributions, index_content_report_content_proxy,
    index_feature_type_content, post_to_instant_articles_api
)
from bulbs.utils.methods import datetime_to_epoch_seconds, get_template_choices
from bulbs.utils import vault
from .managers import ContentManager
from .tasks import update_feature_type_rates

logger = logging.getLogger(__name__)


TEMPLATE_CHOICES = get_template_choices()


class ElasticsearchImageField(field.Integer):

    def to_es(self, data):
        return data.id


class TagManager(PolymorphicManager, IndexableManager):
    pass


class Tag(PolymorphicModel, Indexable):

    """Model for tagging up Content.
    """

    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)

    class Mapping:
        name = field.String(
            analyzer="autocomplete", fields={"raw": field.String(index="not_analyzed")}
        )

    search_objects = TagManager()

    def __unicode__(self):
        """unicode friendly name
        """
        return '%s: %s' % (self.__class__.__name__, self.name)

    def save(self, *args, **kwargs):
        """sets the `slug` values as the name

        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `super.save()`
        """
        self.slug = slugify(self.name)[:50]
        return super(Tag, self).save(*args, **kwargs)

    def count(self):
        """gets the count of instances saved in the cache

        :return: `int`
        """
        return TagCache.count(self.slug)

    @classmethod
    def get_serializer_class(cls):
        """gets the serializer class for the model

        :return: `rest_framework.serializers.Serializer`
        """
        from .serializers import TagSerializer
        return TagSerializer


class FeatureType(Indexable):

    """
    special model for featured content
    """

    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)
    instant_article = models.BooleanField(default=False)

    def __init__(self, *args, **kwargs):
        super(FeatureType, self).__init__(*args, **kwargs)
        # Reference for state change on save.
        self._db_instant_article = self.instant_article
        self._db_pk = self.pk

    class Mapping:
        name = field.String(
            analyzer="autocomplete", fields={"raw": field.String(index="not_analyzed")}
        )
        slug = field.String(index="not_analyzed")

    def __unicode__(self):
        """unicode friendly name
        """
        return self.name

    @property
    def is_new(self):
        return (
            self.pk != self._db_pk or
            self.pk is None and self._db_pk is None
        )

    def save(self, *args, **kwargs):
        """sets the `slug` values as the name

        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `super.save()`
        """
        if self.slug is None or self.slug == "":
            self.slug = slugify(self.name)

        feature_type = super(FeatureType, self).save(*args, **kwargs)

        if self.instant_article_is_dirty:
            index_feature_type_content.delay(self.pk)

        self._db_instant_article = self.instant_article

        # Run all behaviors for `create`
        if self.is_new:
            update_feature_type_rates.delay(self.pk)

        return feature_type

    @property
    def instant_article_is_dirty(self):
        return bool(self.instant_article != self._db_instant_article)


class TemplateType(models.Model):

    """
    Template type for Content.
    """

    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)
    content_type = models.ForeignKey(ContentType)


def _percolate(index, doc_type, content_id, body):
    try:
        results = Content.search_objects.client.percolate(index=index,
                                                          doc_type=doc_type,
                                                          id=content_id,
                                                          body=body)
    except TransportError:
        logger.exception('Percolator error: Content ID %s, %s', content_id, body)
        return []

    # Log any errors (but still try to return any results)
    if results.get('_shards', {}).get('failures'):
        logger.error('Elasticsearch error: %s', results.get('_shards'))

    if results["total"] > 0:
        return results['matches']
    else:
        return []


class Content(PolymorphicModel, Indexable):

    """
    The base content model from which all other content derives.
    """

    published = models.DateTimeField(blank=True, null=True)
    last_modified = models.DateTimeField(auto_now=True)
    title = models.CharField(max_length=512)
    slug = models.SlugField(blank=True, default='')
    template_type = models.ForeignKey(TemplateType, blank=True, null=True)
    description = models.TextField(max_length=1024, blank=True, default="")
    # field used if thumbnail has been manually overridden
    thumbnail_override = ImageField(null=True, blank=True, editable=False)
    # attribution
    authors = models.ManyToManyField(settings.AUTH_USER_MODEL)
    feature_type = models.ForeignKey(FeatureType, null=True, blank=True)
    subhead = models.CharField(max_length=255, blank=True, default="")
    # tagging
    tags = models.ManyToManyField(Tag, blank=True)
    # Should this item be indexed?
    indexed = models.BooleanField(default=True)
    # Is this content always fresh / relevant?
    evergreen = models.BooleanField(default=False)
    # Is this a read only model? (i.e. from elasticsearch)
    _readonly = False
    instant_article = models.BooleanField(default=False)
    # Tunic Campaign ID
    # NOTE: Don't want to accidentally overwrite derived model campaign_id fields during migration
    # Will rename to "campaign_id" after migration
    tunic_campaign_id = models.IntegerField(blank=True, null=True, default=None)
    # Custom template choice. Configured via BULBS_TEMPLATE_CHOICE
    template_choice = models.IntegerField(default=0, choices=TEMPLATE_CHOICES)
    # Facebook Instant Article ID
    instant_article_id = models.BigIntegerField(blank=True, null=True, default=None)

    # custom ES manager
    search_objects = ContentManager()

    class Meta:
        permissions = (
            ("publish_own_content", "Can publish their own content"),
            ("publish_content", "Can publish content"),
            ("promote_content", "Can promote content"),
        )

    class Mapping:
        pk = field.Integer()
        title = field.String(analyzer="snowball", _boost=2.0)
        slug = field.String(index="not_analyzed")
        status = field.String(index="not_analyzed")
        thumbnail_override = ElasticsearchImageField()

    def __unicode__(self):
        """unicode friendly name
        """
        return '%s: %s' % (self.__class__.__name__, self.title)

    @property
    def es_type(self):
        return "{}_{}".format(
            self._meta.app_label, self._meta.model_name.replace("_elasticsearchresult", "")
        )

    @property
    def type(self):
        # TODO: This is to be removed, but some sites rely on it's existence currently.
        return self.es_type

    @property
    def thumbnail(self):
        """Read-only attribute that provides the value of the thumbnail to display.
        """
        # check if there is a valid thumbnail override
        if self.thumbnail_override.id is not None:
            return self.thumbnail_override

        # otherwise, just try to grab the first image
        first_image = self.first_image
        if first_image is not None:
            return first_image

        # no override for thumbnail and no non-none image field, just return override,
        # which is a blank image field
        return self.thumbnail_override

    @property
    def first_image(self):
        """Ready-only attribute that provides the value of the first non-none image that's
        not the thumbnail override field.
        """
        # loop through image fields and grab the first non-none one
        for model_field in self._meta.fields:
            if isinstance(model_field, ImageField):
                if model_field.name is not 'thumbnail_override':
                    field_value = getattr(self, model_field.name)
                    if field_value.id is not None:
                        return field_value

        # no non-none images, return None
        return None

    def get_absolute_url(self):
        """produces a url to link directly to this instance, given the URL config

        :return: `str`
        """
        try:
            url = reverse("content-detail-view", kwargs={"pk": self.pk, "slug": self.slug})
        except NoReverseMatch:
            url = None
        return url

    def get_status(self):
        """Returns a string representing the status of this item.
        By default, this is one of "draft", "scheduled" or "published".

        :return: `str`
        """
        if self.published:
            return "final"  # The published time has been set
        return "draft"  # No published time has been set
    status = property(get_status)

    def get_targeting(self):
        data = {
            "dfp_feature": getattr(self.feature_type, "slug", None),
            "dfp_contentid": self.pk,
            "dfp_pagetype": self.__class__.__name__.lower(),
            "dfp_slug": self.slug,
            "dfp_evergreen": self.evergreen,
            "dfp_title": strip_tags(self.title),
            "dfp_site": getattr(settings, "DFP_SITE", None)
        }
        if self.published is not None:
            data["dfp_publishdate"] = self.published.isoformat()
        data["dfp_campaign"] = getattr(self, "campaign", None)
        tags = self.ordered_tags()
        data["dfp_section"] = tags[0].slug if tags else None
        return data

    @property
    def is_published(self):
        """determines if the content is/should be live

        :return: `bool`
        """
        if self.published:
            now = timezone.now()
            if now >= self.published:
                return True
        return False

    def ordered_tags(self):
        """gets the related tags

        :return: `list` of `Tag` instances
        """
        tags = list(self.tags.all())
        return sorted(
            tags,
            key=lambda tag: ((type(tag) != Tag) * 100000) + tag.count(),
            reverse=True
        )

    def build_slug(self):
        """strips tagging from the title

        :return: `str`
        """
        return strip_tags(self.title)

    @property
    def is_indexed(self):
        return self.indexed

    def save(self, *args, **kwargs):
        """creates the slug, queues up for indexing and saves the instance

        :param args: inline arguments (optional)
        :param kwargs: keyword arguments
        :return: `bulbs.content.Content`
        """
        if not self.slug:
            self.slug = slugify(self.build_slug())[:self._meta.get_field("slug").max_length]

        if not self.is_indexed:
            if kwargs is None:
                kwargs = {}
            kwargs["index"] = False
        content = super(Content, self).save(*args, **kwargs)
        index_content_contributions.delay(self.id)
        index_content_report_content_proxy.delay(self.id)
        post_to_instant_articles_api.delay(self.id)
        return content

    @classmethod
    def get_serializer_class(cls):
        """gets the serializer for the class

        :return: `rest_framework.serializers.Serializer`
        """
        from .serializers import ContentSerializer
        return ContentSerializer

    def get_template_name(self):
        return dict(TEMPLATE_CHOICES).get(self.template_choice)

    def percolate_special_coverage(self, max_size=10, sponsored_only=False):
        """gets list of active, sponsored special coverages containing this content via
        Elasticsearch Percolator (see SpecialCoverage._save_percolator)

        Sorting:
            1) Manually added
            2) Most recent start date
        """

        # Elasticsearch v1.4 percolator range query does not support DateTime range queries
        # (PercolateContext.nowInMillisImpl is not implemented). Once using
        # v1.6+ we can instead compare "start_date/end_date" to python DateTime
        now_epoch = datetime_to_epoch_seconds(timezone.now())

        MANUALLY_ADDED_BOOST = 10
        SPONSORED_BOOST = 100  # Must be order of magnitude higher than "Manual" boost

        # Unsponsored boosting to either lower priority or exclude
        if sponsored_only:
            # Omit unsponsored
            unsponsored_boost = 0
        else:
            # Below sponsored (inverse boost, since we're filtering on "sponsored=False"
            unsponsored_boost = (1.0 / SPONSORED_BOOST)

        # ES v1.4 has more limited percolator capabilities than later
        # implementations. As such, in order to get this to work, we need to
        # sort via scoring_functions, and then manually filter out zero scores.
        sponsored_filter = {
            "query": {
                "function_score": {
                    "functions": [

                        # Boost Recent Special Coverage
                        # Base score is start time
                        # Note: ES 1.4 sorting granularity is poor for times
                        # within 1 hour of each other.
                        {

                            # v1.4 "field_value_factor" does not yet support
                            # "missing" param, and so must filter on whether
                            # "start_date" field exists.
                            "filter": {
                                "exists": {
                                    "field": "start_date",
                                },
                            },
                            "field_value_factor": {
                                "field": "start_date",
                            }
                        },
                        {
                            # Related to above, if "start_date" not found, omit
                            # via zero score.
                            "filter": {
                                "not": {
                                    "exists": {
                                        "field": "start_date",
                                    },
                                },
                            },
                            "weight": 0,
                        },


                        # Ignore non-special-coverage percolator entries
                        {
                            "filter": {
                                "not": {
                                    "prefix": {"_id": "specialcoverage"},
                                },
                            },
                            "weight": 0,
                        },

                        # Boost Manually Added Content
                        {
                            "filter": {
                                "terms": {
                                    "included_ids": [self.id],
                                }
                            },
                            "weight": MANUALLY_ADDED_BOOST,
                        },
                        # Penalize Inactive (Zero Score Will be Omitted)
                        {
                            "filter": {
                                "or": [
                                    {
                                        "range": {
                                            "start_date_epoch": {
                                                "gt": now_epoch,
                                            },
                                        }
                                    },
                                    {
                                        "range": {
                                            "end_date_epoch": {
                                                "lte": now_epoch,
                                            },
                                        }
                                    },
                                ],
                            },
                            "weight": 0,
                        },
                        # Penalize Unsponsored (will either exclude or lower
                        # based on "sponsored_only" flag)
                        {
                            "filter": {
                                "term": {
                                    "sponsored": False,
                                }
                            },
                            "weight": unsponsored_boost,
                        },
                    ],
                },
            },

            "sort": "_score",  # The only sort method supported by ES v1.4 percolator
            "size": max_size,  # Required for sort
        }

        results = _percolate(index=self.mapping.index,
                             doc_type=self.mapping.doc_type,
                             content_id=self.id,
                             body=sponsored_filter)

        return [r["_id"] for r in results
                # Zero score used to omit results via scoring function (ex: inactive)
                if r['_score'] > 0]


class LogEntryManager(models.Manager):

    """
    provides additional manager methods for `bulbs.content.LogEntry` model
    """

    def log(self, user, content, message):
        """creates a new log record

        :param user: user
        :param content: content instance
        :param message: change information
        """
        return self.create(
            user=user,
            content_type=ContentType.objects.get_for_model(content),
            object_id=content.pk,
            change_message=message
        )


class LogEntry(models.Model):

    """
    log entries for changes to content
    """

    action_time = models.DateTimeField("action time", auto_now=True)
    content_type = models.ForeignKey(ContentType, blank=True, null=True, related_name="change_logs")
    object_id = models.TextField("object id", blank=True, null=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, related_name="change_logs")
    change_message = models.TextField("change message", blank=True)
    # custom manager
    objects = LogEntryManager()

    class Meta:
        ordering = ("-action_time",)


class ObfuscatedUrlInfo(Model):

    """
    Stores info used for obfuscated urls of unpublished content.
    """

    content = models.ForeignKey(Content)
    create_date = models.DateTimeField()
    expire_date = models.DateTimeField()
    url_uuid = models.CharField(max_length=32, unique=True, editable=False)

    def save(self, *args, **kwargs):
        """sets uuid for url

        :param args: inline arguments (optional)
        :param kwargs: keyword arguments (optional)
        :return: `super.save()`
        """
        if not self.id:  # this is a totally new instance, create uuid value
            self.url_uuid = str(uuid.uuid4()).replace("-", "")
        super(ObfuscatedUrlInfo, self).save(*args, **kwargs)


##
# signal functions


def content_deleted(sender, instance=None, **kwargs):
    """removes content from the ES index when deleted from DB
    """
    if getattr(instance, "_index", True):
        cls = instance.get_real_instance_class()
        index = cls.search_objects.mapping.index
        doc_type = cls.search_objects.mapping.doc_type

        cls.search_objects.client.delete(index, doc_type, instance.id, ignore=[404])


def delete_from_instant_article_api(sender, instance=None, **kwargs):
    if getattr(settings, 'FACEBOOK_POST_TO_IA', False):
        fb_api_url = getattr(settings, 'FACEBOOK_API_BASE_URL', None)
        fb_token_path = getattr(settings, 'FACEBOOK_TOKEN_VAULT_PATH', None)

        if not fb_api_url or not fb_token_path:
            logger.error('''
                Error in Django Settings.\n
                FACEBOOK_API_BASE_URL: {0}\n
                FACEBOOK_TOKEN_VAULT_PATH: {1}'''.format(fb_api_url,
                                                         fb_token_path))
            return

        fb_access_token = vault.read(fb_token_path).get('authtoken')
        if fb_access_token is None:
            logger.error('Missing FB Auth Token in Vault.\n')
            return

        delete = requests.delete('{0}/{1}?access_token={2}'.format(
            fb_api_url,
            instance.instant_article_id,
            fb_access_token
        ))

        if not delete.ok:
            logger.error('''
                Error in deleting Instant Article.\n
                Content ID: {0}\n
                IA ID: {1}\n
                Status Code: {2}
                Request: {3}'''.format(instance.id,
                                       instance.instant_article_id,
                                       delete.status_code,
                                       delete.__dict__))
        else:
            status = delete.json().get('success')
            if bool(status) is not True:
                logger.error('''
                    Error in deleting Instant Article.\n
                    Content ID: {0}\n
                    IA ID: {1}\n
                    Error: {2}'''.format(instance.id, instance.instant_article_id, delete.json()))

##
# signal hooks

models.signals.pre_delete.connect(content_deleted, Content)
models.signals.pre_delete.connect(delete_from_instant_article_api, Content)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('content', '0007_content_evergreen'),
    ]

    operations = [
        migrations.CreateModel(
            name='Answer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('sodahead_answer_id', models.CharField(default=b'', max_length=20, blank=True)),
                ('answer_text', models.TextField(default=b'', blank=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Poll',
            fields=[
                ('content_ptr', models.OneToOneField(parent_link=True, auto_created=True, primary_key=True, serialize=False, to='content.Content')),
                ('question_text', models.TextField(default=b'', blank=True)),
                ('sodahead_id', models.CharField(default=b'', max_length=20, blank=True)),
                ('last_answer_index', models.IntegerField(default=0)),
                ('end_date', models.DateTimeField(default=None, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('content.content',),
        ),
        migrations.AddField(
            model_name='answer',
            name='poll',
            field=models.ForeignKey(related_name='answers', to='poll.Poll'),
        ),
    ]



from rest_framework.metadata import SimpleMetadata

from bulbs.infographics.metadata import InfographicMetadata
from bulbs.infographics.utils import get_infographics_serializer
from bulbs.super_features.metadata import BaseSuperFeatureMetadata
from bulbs.super_features.utils import get_superfeature_serializer


SUPERFEATURE_SERIALIZER = get_superfeature_serializer()
INFOGRAPHICS_SERIALIZER = get_infographics_serializer()


class PolymorphicContentMetadata(SimpleMetadata):

    serializer_lookup = {
        INFOGRAPHICS_SERIALIZER: InfographicMetadata(),
        SUPERFEATURE_SERIALIZER: BaseSuperFeatureMetadata()
    }

    def determine_metadata(self, request, view):
        if hasattr(view, "get_serializer_class"):
            serializer_class = view.get_serializer_class()
            metadata = self.serializer_lookup.get(serializer_class, None)
            if metadata:
                return metadata.determine_metadata(request, view)
        # TODO: Spike out why we included this generic (and bad) response.
        return {"status": "ok"}



from django.core.management.base import BaseCommand
from bulbs.special_coverage.models import SpecialCoverage


class Command(BaseCommand):
    help = 'Indexes all Special Coverage sections'

    def handle(self, *args, **options):
        for sc in SpecialCoverage.objects.filter(active=True):
            if sc.query:
                sc._save_percolator()



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.cache import add_never_cache_headers
from django.views.decorators.cache import cache_control

from bulbs.content.views import BaseContentDetailView
from bulbs.special_coverage.models import SpecialCoverage
from bulbs.utils.methods import redirect_unpublished_to_login_or_404


class SpecialCoverageView(BaseContentDetailView):
    redirect_correct_path = False

    def get_template_names(self):
        template_names = ["special_coverage/landing.html"]
        template_names.insert(0, getattr(self.special_coverage, "custom_template_name", ""))
        return template_names

    def get(self, request, *args, **kwargs):
        response = super(SpecialCoverageView, self).get(request, *args, **kwargs)

        # Extra unpublished check on Special Coverage activation (BaseContentDetailView only checks
        # first piece of content).
        if not self.special_coverage.is_active:
            if self.show_published_only():
                raise Http404("Special Coverage does not exist.")
            elif not request.user.is_staff:
                return redirect_unpublished_to_login_or_404(request=request,
                                                            next_url=request.get_full_path())

            # Never cache unpublished content
            add_never_cache_headers(response)

        return response

    def get_object(self, *args, **kwargs):
        self.special_coverage = get_object_or_404(SpecialCoverage, slug=self.kwargs.get("slug"))

        qs = self.special_coverage.get_content(published=self.show_published_only()).full()
        if qs.count() == 0:
            raise Http404("No Content available in content list")

        return qs[0]

    def get_context_data(self, *args, **kwargs):
        context = super(SpecialCoverageView, self).get_context_data()
        context["content_list"] = self.special_coverage.get_content(
            published=self.show_published_only()
        )
        if hasattr(self.object, "get_reading_list"):
            context["reading_list"] = self.object.get_reading_list()
        context["special_coverage"] = self.special_coverage
        context["targeting"] = {}

        try:
            context["current_video"] = self.special_coverage.videos[0]
        except IndexError:
            context["current_video"] = None

        if self.special_coverage:
            context["targeting"]["dfp_specialcoverage"] = self.special_coverage.slug
            if self.special_coverage.tunic_campaign_id:
                context["targeting"]["dfp_campaign_id"] = self.special_coverage.tunic_campaign_id
        return context

    def show_published_only(self):
        """
        Returns True if `full_preview` is not a query_parameter.
        Used to determine unpublished preview state.
        """
        return bool("full_preview" not in self.request.GET)


class SpecialCoverageVideoView(SpecialCoverageView):

    def get_context_data(self, *args, **kwargs):
        context = super(SpecialCoverageVideoView, self).get_context_data()

        video_id = int(self.kwargs.get('video_id'))
        if video_id not in self.special_coverage.videos:
            raise Http404('Video with id={} not in SpecialCoverage'.format(video_id))

        context['current_video'] = video_id

        return context


special_coverage = cache_control(max_age=600)(SpecialCoverageView.as_view())
special_coverage_video = cache_control(max_age=600)(SpecialCoverageVideoView.as_view())
